/*
** JavaTalkClient.java
**
** CS549 Spring 96, Oregon State University
** 
** Jean-Guy Speton, Maharajapuram Balakrishnan
*/

import java.applet.*;
import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;

/**
 * @version 1.01, 06 Mar 97
 */
public class JavaTalkClient extends Applet
{
  private static final String VERSION_STRING =
    "JavaTalk 1.01 Copyright (C) 1996-7.  All rights reserved.";
  
  private int port = 8421;   // Default, unless changed by parameter.
  
  protected Socket socket = null;
  protected DataOutputStream out = null;
  private JavaTalkClientReader reader;
  
  protected TextField name, host;
  private Button connect, disconnect;
  private Panel user_list_panel;
  private GridBagLayout user_list_layout;
  protected List user_list;
  protected Vector user_vector;
  private String default_user_name = "Guest";
  protected TextArea text_area;
  protected TextField text_field;

  protected JavaTalkClientGfx gfx_frame;

  public void init()
  {
    System.out.println(VERSION_STRING);
    
    String portParam = getParameter("port");
    if (portParam != null) {
      try {
	port = Integer.parseInt(portParam);
      }
      catch (NumberFormatException nfe) {
	// use default port
	System.err.println("Illegal value for 'port' parameter: " +
			   portParam);
      }
    }

    this.setLayout(new GridBagLayout());

    // add the top row of labels, text fields and buttons
    constrain(this, new Label("Name:"), 0, 0, 1, 1,
	      GridBagConstraints.NONE, GridBagConstraints.CENTER,
	      0.0, 0.0, 2, 2, 2, 2);

    name = new TextField(default_user_name, 16);
    constrain(this, name, 1, 0, 1, 1, GridBagConstraints.HORIZONTAL,
	      GridBagConstraints.CENTER, 1.0, 0.0, 2, 2, 2, 2);

    connect = new Button("Connect");
    connect.setBackground(Color.lightGray);
    constrain(this, connect, 2, 0, 1, 1, GridBagConstraints.BOTH,
	      GridBagConstraints.CENTER, 0.5, 0.0, 2, 2, 2, 2);

    // text area which contains message sent by self and other
    // clients
    text_area = new TextArea("");
    text_area.setEditable(false);
    constrain(this, text_area, 0, 1, 2, 1, GridBagConstraints.BOTH,
	      GridBagConstraints.CENTER, 1.0, 1.0, 2, 2, 2, 2);

    // list of users currently connected
    user_list = new List();
    constrain(this, user_list, 2, 1, 1, 1, GridBagConstraints.BOTH,
	      GridBagConstraints.NORTHWEST, 1.0, 1.0, 2, 2, 2, 2);

    // The user's message to be sent to all.
    text_field = new TextField("");
    text_field.setEditable(false);
    constrain(this, text_field, 0, 2, 2, 1, GridBagConstraints.BOTH,
	      GridBagConstraints.CENTER, 1.0, 0.0, 2, 2, 2, 2);

    //    constrain(this, new Label("Host:"), 3, 3, 1, 1);
    host = new TextField("");
    host.setEditable(false);
    constrain(this, host, 2, 2, 1, 1, GridBagConstraints.BOTH,
	      GridBagConstraints.CENTER, 1.0, 0.0, 2, 2, 2, 2);

    user_vector = new Vector();
  }

  public void stop()
  {
    if (socket != null) {
      doDisconnect();
    }
  }
  
  public boolean handleEvent(Event event)
  {
    switch (event.id) {

    case Event.LIST_SELECT:
      if (event.target == user_list) {
	JavaTalkUser user;
	user = (JavaTalkUser)
	  user_vector.elementAt(((Integer)event.arg).intValue());
	host.setText(user.host + ":" + user.port);
	return true;
      }
      break;

    case Event.LIST_DESELECT:
      if (event.target == user_list) {
	host.setText("");
	return true;
      }
      break;

    }
    return super.handleEvent(event);
  }

  public boolean action(Event event, Object arg)
  {
    if (event.target == connect) {
      if (((String)event.arg).equals("Connect"))
	doConnect();
      else
	doDisconnect();
      return true;
    }
    else if (event.target == text_field) {
      if (text_field.getText().length() > 0) {
	String selected = user_list.getSelectedItem();

	// broadcast to all if no target selected
	if (selected == null) {
	  synchronized (out) {
	    try {
	      out.writeInt(JavaTalkMessage.TEXT);
	      out.writeBytes(name.getText() + "\n");
	      out.writeBytes(text_field.getText() + "\n");
	      text_field.setText("");
	    }
	    catch (IOException e) {
	      System.err.println("Text: " + e);
	    }
	  }
	}
	else {   // send private message to selected user
	  synchronized (out) {
	    try {
	      out.writeInt(JavaTalkMessage.PRIV_TEXT);
	      out.writeBytes(selected + "\n");
	      out.writeBytes(name.getText() + "\n");
	      out.writeBytes(text_field.getText() + "\n");
	      text_area.appendText("--> " + selected + ": " +
				   text_field.getText() + "\n");
	      text_field.setText("");
	    }
	    catch (IOException e) {
	      System.err.println("Private text: " + e);
	    }
	  }
	}
      }
      return true;
    }
    else
      return super.action(event, arg);
  }

  private void doConnect()
  {
    try {
      socket = new Socket(getCodeBase().getHost(), port);
      out = new DataOutputStream(socket.getOutputStream());
      out.writeBytes(name.getText() + "\n");
      reader = new JavaTalkClientReader(this);
      reader.start();
    }
    catch (IOException e) {
      showStatus("Exception: " + e);
      return;
    }
    
    // create the graphics canvas window, add GUI elements, display
    gfx_frame = new JavaTalkClientGfx(this);
    gfx_frame.pack();
    gfx_frame.show();

    // update applet buttons
    connect.setLabel("Disconnect");

    // disallow name changes while connected
    name.setEditable(false);
    // allow message entry
    text_field.setEditable(true);
  }

  private void doDisconnect()
  {
    // close socket connection
    try {
      out.writeInt(JavaTalkMessage.DISCONNECT);
    }
    catch (IOException e) {
      showStatus("DISCONNECT msg failed: " + e);
    }

    try {
      socket.close();
    }
    catch (IOException e) {
      showStatus("socket.close() failed: " + e);
    }
    socket = null;

    // clear list
    user_list.clear();
    host.setText("");

    // remove text and graphic windows
    gfx_frame.dispose();
    
    // update applet buttons
    connect.setLabel("Connect");

    // allow name changes while disconnected
    name.setEditable(true);
    // disallow messages while disconnected
    text_field.setEditable(false);
    text_area.setText("");
  }

  protected void sendGfxMessage(int mode, int color, int x1, int y1,
				int x2, int y2)
  {
    synchronized (out) {
      try {
	out.writeInt(JavaTalkMessage.GFX);
	out.writeInt(mode);
	out.writeInt(color);
	out.writeInt(x1);
	out.writeInt(y1);
	out.writeInt(x2);
	out.writeInt(y2);
      }
      catch (IOException e) {
	System.err.println("sendGfxMessage: " + e);
      }
    }
  }

  // Simplifies using GridBagContraints objects.
  // Mercilessly lifted from _Java in a Nutshell_.
  public void constrain(Container container, Component component,
			int grid_x, int grid_y,
			int grid_width, int grid_height,
			int fill, int anchor, double weight_x, double weight_y,
			int top, int left, int bottom, int right)
  {
    GridBagConstraints c = new GridBagConstraints();
    c.gridx = grid_x; c.gridy = grid_y;
    c.gridwidth = grid_width; c.gridheight = grid_height;
    c.fill = fill; c.anchor = anchor;
    c.weightx = weight_x; c.weighty = weight_y;
    if (top+bottom+left+right > 0)
      c.insets = new Insets(top, left, bottom, right);

    ((GridBagLayout)container.getLayout()).setConstraints(component, c);
    container.add(component);
  }

  public String getAppletInfo()
  {
    return "JavaTalk 1.0 written by M. Balakrishnan and J. Speton";
  }

}

/*
 * @author Maharajapuram Balakrishnan
 * @author Jean-Guy Speton
 * @see    JavaTalkClient
 *
 * The Reader thread sits on the client's socket and waits for messages
 * to arrive from the server, acting appropriately when such a message
 * is received.
 */
class JavaTalkClientReader extends Thread
{
  JavaTalkClient client;

  public JavaTalkClientReader(JavaTalkClient c)
  {
    super("JavaTalkClient Reader");
    this.client = c;
  }

  public void run()
  {
    DataInputStream in = null;

    try {
      in = new DataInputStream(client.socket.getInputStream());

      // loop now, waiting for messages from the server
      while(true) {
	int msgType;
	JavaTalkUser user;
	
	msgType = in.readInt();
	switch (msgType) {

	case JavaTalkMessage.ADD_USER:
	case JavaTalkMessage.NEW_USER:
	  user = new JavaTalkUser();
	  user.name = in.readLine();
	  user.host = in.readLine();
	  user.port = in.readInt();
	  client.user_list.addItem(user.name);
	  client.user_vector.addElement(user);
	  if (msgType == JavaTalkMessage.ADD_USER) {
	    synchronized (client.text_area) {
	      client.text_area.appendText("++" + user.name +
					  " has joined the conference.\n");
	    }
	  }
	  break;

	case JavaTalkMessage.DEL_USER:
	  String name = new String(in.readLine());
	  for (int i = 0; i < client.user_list.countItems(); i++) {
	    if (name.equals(client.user_list.getItem(i))) {
	      client.user_list.delItem(i);
	      try {
		client.user_vector.removeElementAt(i);
	      }
	      catch (ArrayIndexOutOfBoundsException e) {}
	      break;
	    }
	  }
	  synchronized (client.text_area) {
	    client.text_area.appendText("--" + name +
					" has left the conference.\n");
	  }
	  break;

	case JavaTalkMessage.TEXT:
	  String username, msg;
	  // read and print text message to text window
	  username = in.readLine();
	  msg = in.readLine();
	  client.text_area.appendText(username + ": "	+ msg + "\n");
	  break;

	case JavaTalkMessage.GFX:
	  int mode, x1, y1, x2, y2;
	  Color color;
	  
	  mode = in.readInt();
	  switch (in.readInt()) {
	  case 0: color = Color.black; break;
	  case 1: color = Color.blue; break;
	  case 2: color = Color.cyan; break;
	  case 3: color = Color.gray; break;
	  case 4: color = Color.green; break;
	  case 5: color = Color.magenta; break;
	  case 6: color = Color.orange; break;
	  case 7: color = Color.pink; break;
	  case 8: color = Color.red; break;
	  case 9: color = Color.yellow; break;
	  default: color = Color.black; break;  // should never happen
	  }
	  x1 = in.readInt();
	  y1 = in.readInt();
	  x2 = in.readInt();
	  y2 = in.readInt();

	  client.gfx_frame.drawGfx(mode, x1, y1, x2, y2, color);

	  break;

	default:
	  System.err.println("Unrecognized message: " + msgType);
	}

	if (msgType == JavaTalkMessage.SHUTDOWN) { 
	  client.showStatus("Server closed connection.");
	  break;
	}
      }
    }
    catch (IOException e) {
      System.err.println("Reader: " + e);
    }
    finally {
      try {
	if (in != null) {
	  in.close();
	}
      }
      catch (IOException e) {
      }
    }
  }
}
